from django.db import models
from django.utils.translation import gettext_lazy as _

from filer.fields import image

from cms.models.pagemodel import Page

from shop import deferred
from shop.models.product import BaseProduct


class BaseProductPage(models.Model, metaclass=deferred.ForeignKeyBuilder):
    """
    ManyToMany relation from the polymorphic Product to the CMS Page.
    This in practice is the category.
    """
    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
    )

    product = deferred.ForeignKey(
        BaseProduct,
        on_delete=models.CASCADE,
    )

    class Meta:
        abstract = True
        unique_together = ['page', 'product']
        verbose_name = _("Category")
        verbose_name_plural = _("Categories")

ProductPageModel = deferred.MaterializedModel(BaseProductPage)


class BaseProductImage(models.Model, metaclass=deferred.ForeignKeyBuilder):
    """
    ManyToMany relation from the polymorphic Product to a set of images.
    """
    image = image.FilerImageField(on_delete=models.CASCADE)

    product = deferred.ForeignKey(
        BaseProduct,
        on_delete=models.CASCADE,
    )

    order = models.SmallIntegerField(default=0)

    class Meta:
        abstract = True
        verbose_name = _("Product Image")
        verbose_name_plural = _("Product Images")
        ordering = ['order']

ProductImageModel = deferred.MaterializedModel(BaseProductImage)



from django.db.models import SET_DEFAULT

from shop import deferred
from shop.models.address import BaseShippingAddress, BaseBillingAddress
from shop.models.cart import BaseCart


class Cart(BaseCart):
    """
    Default materialized model for BaseCart containing common fields
    """
    shipping_address = deferred.ForeignKey(
        BaseShippingAddress,
        on_delete=SET_DEFAULT,
        null=True,
        default=None,
        related_name='+',
    )

    billing_address = deferred.ForeignKey(
        BaseBillingAddress,
        on_delete=SET_DEFAULT,
        null=True,
        default=None,
        related_name='+',
    )



import warnings

from django.urls import include, path
    
from shop.modifiers.pool import cart_modifiers_pool


urlpatterns = []

# For every payment modifier, load the URLs from the associated `payment_provider`.
for modifier in cart_modifiers_pool.get_payment_modifiers():
    try:
        namespace = modifier.payment_provider.namespace
        path_ = '{}/'.format(namespace)
        urls = modifier.payment_provider.get_urls()
        provider_url = path( path_, include((urls, 'url_payment'), namespace=namespace))
        if provider_url.url_patterns != []:
            urlpatterns.append(provider_url)
    except AttributeError as err:
        warnings.warn(err.message)



from django.db.models import PositiveIntegerField
from shop.models import cart


class CartItem(cart.BaseCartItem):
    """Default materialized model for CartItem"""
    quantity = PositiveIntegerField()



from elasticsearch_dsl.analysis import analyzer, token_filter

body_analyzers = {
    'default': analyzer('default_analyzer',
        tokenizer='standard',
        filter=['lowercase', 'stop', 'snowball'],
        char_filter=['html_strip'],
    ),
    'de': analyzer('german_analyzer',
        type='custom',
        tokenizer='standard',
        filter=[
            'lowercase',
            token_filter('asciifolding', type='asciifolding', preserve_original=False),
            token_filter('german_stop', type='stop', language='german'),
            token_filter('german_stemmer', type='snowball', language='german'),
        ],
        char_filter=['html_strip'],
    ),
}



from django.template.loader import select_template
from django.utils import translation

from django_elasticsearch_dsl import fields, Document, Index
from elasticsearch.exceptions import NotFoundError

from shop.conf import app_settings
from shop.models.product import ProductModel
from shop.search.analyzers import body_analyzers


class _ProductDocument(Document):
    product_code = fields.KeywordField(
        multi=True,
        boost=3,
    )

    product_name = fields.TextField(
        boost=2,
    )

    product_type = fields.TextField()

    class Django:
        model = ProductModel
        fields = ['id']
        ignore_signals = True  # performed by ProductModel.update_search_index()
        queryset_pagination = 499  # let DRF do the pagination

    def __str__(self):
        return "{} {}: {}".format(self.product_type, self.id, self.product_name)

    def get_queryset(self):
        queryset = super().get_queryset()
        return queryset.filter(active=True)

    def prepare_product_code(self, instance):
        """
        Create a list of textual representation for product codes.
        """
        has_valid_product_code = lambda obj: isinstance(getattr(obj, 'product_code', None), str)
        variants = instance.get_product_variants()
        product_codes = [v.product_code for v in variants if has_valid_product_code(v)]
        if has_valid_product_code(instance):
            product_codes.append(instance.product_code)
        return product_codes

    def prepare_body(self, instance):
        """
        Create a textual representation of the product's instance to be used by Elasticsearch for
        creating a full text search index.
        """
        app_label = instance._meta.app_label.lower()
        params = [
            (app_label, instance.product_model),
            (app_label, 'product'),
            ('shop', 'product'),
        ]
        template = select_template(['{0}/search/indexes/{1}.txt'.format(*p) for p in params])
        body = template.render({'product': instance})
        return body

    def update(self, thing, refresh=None, action='index', parallel=False, **kwargs):
        if isinstance(thing, ProductModel._materialized_model) and thing.active is False:
            try:
                doc = self.get(id=thing.id)
            except NotFoundError:
                pass
            else:
                doc.delete()
        else:
            if self._language:
                with translation.override(self._language):
                    super().update(thing, refresh=None, action='index', parallel=False, **kwargs)
            else:
                super().update(thing, refresh=None, action='index', parallel=False, **kwargs)


class ProductDocument:
    """
    Factory for building an elasticsearch-dsl Document class. This class
    """
    def __new__(cls, language=None, settings=None):
        index_name_parts = [app_settings.SHOP_APP_LABEL]
        if language:
            index_name_parts.append(language.lower())
            doc_name = 'ProductDocument{}'.format(language.title())
            analyzer = body_analyzers.get(language, body_analyzers['default'])
        else:
            doc_name = 'ProductDocument'
            analyzer = body_analyzers['default']
        index_name_parts.append('products')
        products_index = Index('.'.join(index_name_parts))
        if settings:
            products_index.settings(**settings)
        attrs = {'_language': language, 'body': fields.TextField(analyzer=analyzer)}
        doc_class = type(doc_name, (_ProductDocument,), attrs)
        products_index.document(doc_class)
        return doc_class



from django.forms import fields, widgets
from django.utils.translation import gettext_lazy as _, ngettext_lazy
from django.utils.text import Truncator
from django.utils.html import format_html
from django.forms.fields import IntegerField
from django.template.loader import select_template
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.forms import ManageChildrenFormMixin
from cmsplugin_cascade.widgets import NumberInputWidget
from cmsplugin_cascade.plugin_base import TransparentWrapper, TransparentContainer
from shop.conf import app_settings
from shop.cascade.plugin_base import ShopPluginBase


class ProcessBarFormMixin(ManageChildrenFormMixin, EntangledModelFormMixin):
    num_children = IntegerField(
        min_value=1,
        initial=1,
        widget=NumberInputWidget(attrs={'size': '3', 'style': 'width: 5em;'}),
        label=_("Steps"),
        help_text=_("Number of steps for this proceed bar."))

    class Meta:
        untangled_fields = ['num_children']


class ProcessBarPlugin(TransparentWrapper, ShopPluginBase):
    name = _("Process Bar")
    form = ProcessBarFormMixin
    parent_classes = ('BootstrapColumnPlugin',)
    direct_child_classes = ('ProcessStepPlugin',)
    require_parent = True
    allow_children = True

    @classmethod
    def get_identifier(cls, instance):
        identifier = super().get_identifier(instance)
        num_cols = instance.get_children().count()
        content = ngettext_lazy('with {} page', 'with {} pages', num_cols).format(num_cols)
        return format_html('{0}{1}', identifier, content)

    def get_render_template(self, context, instance, placeholder):
        template_names = [
            '{}/checkout/process-bar.html'.format(app_settings.APP_LABEL),
            'shop/checkout/process-bar.html',
        ]
        return select_template(template_names)

    def render(self, context, instance, placeholder):
        self.super(ProcessBarPlugin, self).render(context, instance, placeholder)
        num_children = instance.get_num_children()
        if num_children > 0:
            context['step_css_width'] = '{:3.2f}%'.format(100. / num_children)
        return context

    def save_model(self, request, obj, form, change):
        wanted_children = int(form.cleaned_data.get('num_children'))
        super().save_model(request, obj, form, change)
        self.extend_children(obj, wanted_children, ProcessStepPlugin)

plugin_pool.register_plugin(ProcessBarPlugin)


class ProcessStepFormMixin(EntangledModelFormMixin):
    step_title = fields.CharField(
        widget=widgets.TextInput(attrs={'size': 50}),
        label=_("Step Title"),
        required=False,
    )

    class Meta:
        entangled_fields = {'glossary': ['step_title']}


class ProcessStepPlugin(TransparentContainer, ShopPluginBase):
    name = _("Process Step")
    direct_parent_classes = parent_classes = ['ProcessBarPlugin']
    require_parent = True
    allow_children = True
    alien_child_classes = True
    form = ProcessStepFormMixin
    render_template = 'cascade/generic/wrapper.html'

    @classmethod
    def get_identifier(cls, obj):
        identifier = super().get_identifier(obj)
        content = obj.glossary.get('step_title', '')
        if content:
            content = Truncator(content).words(3, truncate=' ...')
        else:
            content = obj.get_position_in_placeholder()
        return format_html('{0}{1}', identifier, content)

plugin_pool.register_plugin(ProcessStepPlugin)



from django.core.exceptions import ImproperlyConfigured
from django.forms import fields, widgets
from django.template import engines, TemplateDoesNotExist
from django.template.loader import select_template
from django.utils.html import mark_safe
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin
from cms.apphook_pool import apphook_pool
from cms.plugin_pool import plugin_pool
from shop.conf import app_settings
from shop.cascade.plugin_base import ShopPluginBase


class BreadcrumbPluginForm(EntangledModelFormMixin):
    CHOICES = [
        ('default', _("Default")),
        ('soft-root', _("With “Soft-Root”")),
        ('catalog', _("With Catalog Count")),
    ]

    render_type = fields.ChoiceField(
        choices=CHOICES,
        widget=widgets.RadioSelect,
        label=_("Render as"),
        initial='default',
        help_text=_("Render an alternative Breadcrumb"),
    )

    class Meta:
        entangled_fields = {'glossary': ['render_type']}


class BreadcrumbPlugin(ShopPluginBase):
    name = _("Breadcrumb")
    parent_classes = []
    allow_children = None
    form = BreadcrumbPluginForm

    @classmethod
    def get_identifier(cls, instance):
        render_type = instance.glossary.get('render_type')
        return mark_safe(dict(cls.form.CHOICES).get(render_type, ''))

    def get_render_template(self, context, instance, placeholder):
        render_type = instance.glossary.get('render_type')
        try:
            return select_template([
                '{}/breadcrumb/{}.html'.format(app_settings.APP_LABEL, render_type),
                'shop/breadcrumb/{}.html'.format(render_type),
            ])
        except TemplateDoesNotExist:
            return engines['django'].from_string('<!-- empty breadcrumb -->')

    def get_use_cache(self, context, instance, placeholder):
        try:
            app = apphook_pool.get_apphook(instance.page.application_urls)
            return app.cache_placeholders
        except (AttributeError, ImproperlyConfigured):
            return super().get_use_cache(context, instance, placeholder)

plugin_pool.register_plugin(BreadcrumbPlugin)



from django.conf.urls import include, url
from shop.urls import rest_api
from shop.urls import auth
from shop.urls import payment


app_name = 'shop'

urlpatterns = [
    url(r'^api/', include(rest_api)),
    url(r'^auth/', include(auth)),
    url(r'^payment/', include(payment)),
]



from django.forms import widgets
from django.core.exceptions import ValidationError
from django.template import engines
from django.template.loader import select_template
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.bootstrap4.buttons import BootstrapButtonMixin, ButtonFormMixin
from cmsplugin_cascade.icon.forms import IconFormMixin
from cmsplugin_cascade.plugin_base import TransparentWrapper
from djng.forms import fields, NgModelFormMixin
from djng.styling.bootstrap3.forms import Bootstrap3Form
from shop.cascade.extensions import ShopExtendableMixin, LeftRightExtensionMixin
from shop.cascade.plugin_base import ShopPluginBase
from shop.conf import app_settings


class ShopOrderViewsFormMixin(EntangledModelFormMixin):
    def clean(self):
        cleaned_data = super().clean()
        if self.instance.page and self.instance.page.application_urls != 'OrderApp':
            msg = "This plugin only makes sense if used on a CMS page with an application of type 'OrderApp'."
            raise ValidationError(msg)
        return cleaned_data


class ShopOrderViewsPlugin(LeftRightExtensionMixin, TransparentWrapper, ShopPluginBase):
    name = _("Order Views")
    require_parent = True
    parent_classes = ['BootstrapColumnPlugin']
    allow_children = True
    model_mixins = (ShopExtendableMixin,)
    form = ShopOrderViewsFormMixin
    cache = False

    def get_render_template(self, context, instance, placeholder):
        many = context.get('many')
        if many is True:
            # render Order List View
            return select_template([
                '{}/order/list.html'.format(app_settings.APP_LABEL),
                'shop/order/list.html',
            ])
        if many is False:
            # render Order Detail View
            return select_template([
                '{}/order/detail.html'.format(app_settings.APP_LABEL),
                'shop/order/detail.html',
            ])
        # can happen, if this plugin is abused outside of an OrderView
        alert_msg = '''<div class="alert alert-danger">
        This {} plugin is used on a CMS page without an application of type "View Order".
        </div>'''
        return engines['django'].from_string(alert_msg.format(self.name))

plugin_pool.register_plugin(ShopOrderViewsPlugin)


class OrderButtonForm(ShopOrderViewsFormMixin, IconFormMixin, ButtonFormMixin):
    require_icon = False


class OrderButtonBase(BootstrapButtonMixin, ShopPluginBase):
    parent_classes = ['BootstrapColumnPlugin']
    form = OrderButtonForm

    @classmethod
    def get_identifier(cls, instance):
        return instance.glossary.get('button_content', '')


class ShopReorderButtonPlugin(OrderButtonBase):
    name = _("Reorder Button")

    def get_render_template(self, context, instance, placeholder):
        template_names = [
            '{}/order/reorder-button.html'.format(app_settings.APP_LABEL),
            'shop/order/reorder-button.html',
        ]
        return select_template(template_names)

plugin_pool.register_plugin(ShopReorderButtonPlugin)


class ShopCancelOrderButtonPlugin(OrderButtonBase):
    name = _("Cancel Order Button")

    def get_render_template(self, context, instance, placeholder):
        template_names = [
            '{}/order/cancel-button.html'.format(app_settings.APP_LABEL),
            'shop/order/cancel-button.html',
        ]
        return select_template(template_names)

plugin_pool.register_plugin(ShopCancelOrderButtonPlugin)


class AddendumForm(NgModelFormMixin, Bootstrap3Form):
    annotation = fields.CharField(
        label="",
        widget=widgets.Textarea(attrs={'rows': 2}),
    )


class ShopOrderAddendumFormMixin(OrderButtonForm):
    show_history = fields.BooleanField(
         label=_("Show History"),
         initial=True,
         required=False,
         help_text=_("Show historical annotations."),
    )

    class Meta:
        entangled_fields = {'glossary': ['show_history']}


class ShopOrderAddendumFormPlugin(OrderButtonBase):
    name = _("Order Addendum Form")
    form = ShopOrderAddendumFormMixin

    def get_render_template(self, context, instance, placeholder):
        template_names = [
            '{}/order/addendum-form.html'.format(app_settings.APP_LABEL),
            'shop/order/addendum-form.html',
        ]
        return select_template(template_names)

    def render(self, context, instance, placeholder):
        context = self.super(ShopOrderAddendumFormPlugin, self).render(context, instance, placeholder)
        context.update({
            'addendum_form': AddendumForm(),
            'show_history': instance.glossary.get('show_history', True),
        })
        return context

plugin_pool.register_plugin(ShopOrderAddendumFormPlugin)



from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.contrib import admin
from django.contrib.sites.models import Site
from django.utils.translation import gettext_lazy as _

try:
    from django_elasticsearch_dsl.registries import registry as elasticsearch_registry
except ImportError:
    elasticsearch_registry = type('DocumentRegistry', (), {'get_documents': lambda *args: []})()

from adminsortable2.admin import SortableInlineAdminMixin

from cms.models import Page

from shop.models.related import ProductPageModel, ProductImageModel


class ProductImageInline(SortableInlineAdminMixin, admin.StackedInline):
    model = ProductImageModel
    extra = 1
    ordering = ['order']


def _find_catalog_list_apphook():
    from shop.cms_apphooks import CatalogListCMSApp
    from cms.apphook_pool import apphook_pool

    for name, app in apphook_pool.apps.items():
        if isinstance(app, CatalogListCMSApp):
            return name
    else:
        raise ImproperlyConfigured("You must register a CMS apphook of type `CatalogListCMSApp`.")


class CategoryModelMultipleChoiceField(forms.ModelMultipleChoiceField):
    def label_from_instance(self, obj):
        if Site.objects.count() >=2 :
            page_sitename=str(Site.objects.filter(djangocms_nodes=obj.node_id).first().name)
            return '{} | {}'.format(str(obj), page_sitename)
        else:
            return str(obj)


class CMSPageAsCategoryMixin:
    """
    Add this mixin class to the ModelAdmin class for products wishing to be assigned to djangoCMS
    pages when used as categories.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not hasattr(self.model, 'cms_pages'):
            raise ImproperlyConfigured("Product model requires a field named `cms_pages`")

    def get_fieldsets(self, request, obj=None):
        fieldsets = list(super().get_fieldsets(request, obj=obj))
        fieldsets.append((_("Categories"), {'fields': ('cms_pages',)}),)
        return fieldsets

    def get_fields(self, request, obj=None):
        # In ``get_fieldsets()``, ``cms_pages`` is added, so remove it from ``fields`` to
        # avoid showing it twice.
        fields = list(super().get_fields(request, obj))
        try:
            fields.remove('cms_pages')
        except ValueError:
            pass
        return fields

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        if db_field.name == 'cms_pages':
            # restrict many-to-many field for cms_pages to ProductApp only
            limit_choices_to = {
                'publisher_is_draft': False,
                'application_urls': getattr(self, 'limit_to_cmsapp', _find_catalog_list_apphook()),
            }
            queryset = Page.objects.filter(**limit_choices_to)
            widget = admin.widgets.FilteredSelectMultiple(_("CMS Pages"), False)
            required = not db_field.blank
            field = CategoryModelMultipleChoiceField(queryset=queryset, widget=widget, required=required)
            return field
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def save_related(self, request, form, formsets, change):
        old_cms_pages = form.instance.cms_pages.all()
        new_cms_pages = form.cleaned_data.pop('cms_pages')

        # remove old
        for page in old_cms_pages:
            if page not in new_cms_pages:
                for pp in ProductPageModel.objects.filter(product=form.instance, page=page):
                    pp.delete()

        # add new
        for page in new_cms_pages:
            if page not in old_cms_pages:
                ProductPageModel.objects.create(product=form.instance, page=page)

        return super().save_related(request, form, formsets, change)


class SearchProductIndexMixin:
    """
    If Elasticsearch is used to create a full text search index, add this mixin class to Django's
    ``ModelAdmin`` backend for the corresponding product model.
    """
    def save_model(self, request, product, form, change):
        super().save_model(request, product, form, change)
        if change:
            product.update_search_index()

    def delete_model(self, request, product):
        product.active = False
        product.update_search_index()
        super().delete_model(request, product)


class InvalidateProductCacheMixin:
    """
    If Redis caching is used to create a HTML snippets for product representation, add this mixin
    class to Django's ``ModelAdmin`` backend for the corresponding product model.
    """
    def save_model(self, request, product, form, change):
        if change:
            product.invalidate_cache()
        return super().save_model(request, product, form, change)

    def delete_model(self, request, product):
        product.invalidate_cache()
        super().delete_model(request, product)


class UnitPriceMixin:
    def get_list_display(self, request):
        list_display = list(super().get_list_display(request))
        list_display.append('get_unit_price')
        return list_display

    def get_unit_price(self, obj):
        return str(obj.unit_price)
    get_unit_price.short_description = _("Unit Price")


class CMSPageFilter(admin.SimpleListFilter):
    title = _("Category")
    parameter_name = 'category'

    def lookups(self, request, model_admin):
        limit_choices_to = {
            'publisher_is_draft': False,
            'application_urls': getattr(self, 'limit_to_cmsapp', _find_catalog_list_apphook())
        }
        queryset = Page.objects.filter(**limit_choices_to)
        return [(page.id, page.get_title()) for page in queryset]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(cms_pages__id=self.value())



from django.contrib.admin import StackedInline
from django.forms import fields, widgets
from django.template.loader import select_template
from django.utils.translation import gettext_lazy as _, gettext

from entangled.forms import EntangledModelFormMixin, EntangledModelForm

from cms.plugin_pool import plugin_pool
from cms.utils.compat.dj import is_installed
from cmsplugin_cascade.mixins import WithSortableInlineElementsMixin
from cmsplugin_cascade.models import SortableInlineCascadeElement

from shop.cascade.plugin_base import ShopPluginBase, ProductSelectField
from shop.conf import app_settings
from shop.models.product import ProductModel

if is_installed('adminsortable2'):
    from adminsortable2.admin import SortableInlineAdminMixin
else:
    SortableInlineAdminMixin = type('SortableInlineAdminMixin', (object,), {})


class ShopCatalogPluginForm(EntangledModelFormMixin):
    CHOICES = [
        ('paginator', _("Use Paginator")),
        ('manual', _("Manual Infinite")),
        ('auto', _("Auto Infinite")),
    ]

    pagination = fields.ChoiceField(
        choices=CHOICES,
        widget=widgets.RadioSelect,
        label=_("Pagination"),
        initial='paginator',
        help_text=_("Shall the product list view use a paginator or scroll infinitely?"),
    )

    class Meta:
        entangled_fields = {'glossary': ['pagination']}


class ShopCatalogPlugin(ShopPluginBase):
    name = _("Catalog List View")
    require_parent = True
    form = ShopCatalogPluginForm
    parent_classes = ['BootstrapColumnPlugin', 'SimpleWrapperPlugin']
    cache = False

    def get_render_template(self, context, instance, placeholder):
        templates = []
        if instance.glossary.get('render_template'):
            templates.append(instance.glossary['render_template'])
        templates.extend([
            '{}/catalog/list.html'.format(app_settings.APP_LABEL),
            'shop/catalog/list.html',
        ])
        return select_template(templates)

    def render(self, context, instance, placeholder):
        context['pagination'] = instance.glossary.get('pagination', 'paginator')
        return context

    @classmethod
    def get_identifier(cls, obj):
        pagination = obj.glossary.get('pagination')
        if pagination == 'paginator':
            return gettext("Manual Pagination")
        return gettext("Infinite Scroll")

plugin_pool.register_plugin(ShopCatalogPlugin)


class ShopAddToCartPluginForm(EntangledModelFormMixin):
    use_modal_dialog = fields.BooleanField(
        label=_("Use Modal Dialog"),
        initial=True,
        required=False,
        help_text=_("After adding product to cart, render a modal dialog"),
    )

    class Meta:
        entangled_fields = {'glossary': ['use_modal_dialog']}


class ShopAddToCartPlugin(ShopPluginBase):
    name = _("Add Product to Cart")
    require_parent = True
    form = ShopAddToCartPluginForm
    parent_classes = ['BootstrapColumnPlugin']
    cache = False

    def get_render_template(self, context, instance, placeholder):
        templates = []
        if instance.glossary.get('render_template'):
            templates.append(instance.glossary['render_template'])
        if context['product'].managed_availability():
            template_prefix = 'available-'
        else:
            template_prefix = ''
        templates.extend([
            '{}/catalog/{}product-add2cart.html'.format(app_settings.APP_LABEL, template_prefix),
            'shop/catalog/{}product-add2cart.html'.format(template_prefix),
        ])
        return select_template(templates)

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        context['use_modal_dialog'] = bool(instance.glossary.get('use_modal_dialog', True))
        return context

plugin_pool.register_plugin(ShopAddToCartPlugin)


class ProductGalleryForm(EntangledModelForm):
    order = fields.IntegerField(
        widget=widgets.HiddenInput,
        initial=0,
    )

    product = ProductSelectField(
        required=False,
        label=_("Related Product"),
        help_text=_("Choose related product"),
    )

    class Meta:
        entangled_fields = {'glossary': ['product']}
        untangled_fields = ['order']


class ProductGalleryInline(SortableInlineAdminMixin, StackedInline):
    model = SortableInlineCascadeElement
    form = ProductGalleryForm
    extra = 5
    ordering = ['order']
    verbose_name = _("Product")
    verbose_name_plural = _("Product Gallery")


class ShopProductGallery(WithSortableInlineElementsMixin, ShopPluginBase):
    name = _("Product Gallery")
    require_parent = True
    parent_classes = ('BootstrapColumnPlugin',)
    inlines = (ProductGalleryInline,)
    # until this bug https://github.com/applegrew/django-select2/issues/65 is fixed
    # we hide the a "add row" button and instead use `extra = 5` in ProductGalleryInline

    class Media:
        css = {'all': ('shop/css/admin/product-gallery.css',)}

    def get_render_template(self, context, instance, placeholder):
        templates = []
        if instance.glossary.get('render_template'):
            templates.append(instance.glossary['render_template'])
        templates.extend([
            '{}/catalog/product-gallery.html'.format(app_settings.APP_LABEL),
            'shop/catalog/product-gallery.html',
        ])
        return select_template(templates)

    def render(self, context, instance, placeholder):
        product_ids = []
        for inline in instance.sortinline_elements.all():
            try:
                product_ids.append(inline.glossary['product']['pk'])
            except KeyError:
                pass
        queryset = ProductModel.objects.filter(pk__in=product_ids, active=True)
        serializer_class = app_settings.PRODUCT_SUMMARY_SERIALIZER
        serialized = serializer_class(queryset, many=True, context={'request': context['request']})
        # sort the products according to the order provided by `sortinline_elements`.
        context['products'] = [product for id in product_ids for product in serialized.data if product['id'] == id]
        return context

plugin_pool.register_plugin(ShopProductGallery)



from django.conf import settings
from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from adminsortable2.admin import SortableAdminMixin
from cms.admin.placeholderadmin import PlaceholderAdminMixin, FrontendEditableAdminMixin
from shop.admin.product import CMSPageAsCategoryMixin, UnitPriceMixin, CMSPageFilter
from shop.models.defaults.commodity import Commodity

if settings.USE_I18N:
    from parler.admin import TranslatableAdmin


    @admin.register(Commodity)
    class CommodityAdmin(SortableAdminMixin, TranslatableAdmin, FrontendEditableAdminMixin,
                         PlaceholderAdminMixin, CMSPageAsCategoryMixin, UnitPriceMixin, admin.ModelAdmin):
        fieldsets = [
            (None, {
                'fields': [
                    ('product_name', 'slug'),
                    'caption',
                ]
            }),
            (_("Common Fields"), {
                'fields': [
                    ('product_code', 'quantity'),
                    ('unit_price', 'active'),
                    'show_breadcrumb',
                    'sample_image',
                ],
            }),
        ]
        filter_horizontal = ['cms_pages']
        list_filter = [CMSPageFilter]
        list_display = ['product_name', 'product_code', 'get_unit_price', 'active']

        def get_prepopulated_fields(self, request, obj=None):
            return {
                'slug': ['product_name'],
            }


else:

    @admin.register(Commodity)
    class CommodityAdmin(SortableAdminMixin, FrontendEditableAdminMixin, PlaceholderAdminMixin,
                         CMSPageAsCategoryMixin, UnitPriceMixin, admin.ModelAdmin):
        fields = [
            ('product_name', 'slug'),
            'caption',
            ('product_code', 'quantity'),
            ('unit_price', 'active'),
            'show_breadcrumb',
            'sample_image',
        ]
        filter_horizontal = ['cms_pages']
        list_filter = [CMSPageFilter]
        list_display = ['product_name', 'product_code', 'get_unit_price', 'active']
        prepopulated_fields = {'slug': ['product_name']}



from django.utils.translation import get_language_from_request

from django_elasticsearch_dsl.registries import registry

from shop.models.product import ProductModel


class SearchViewMixin:
    def get_document(self, language):
        documents = registry.get_documents([ProductModel])
        try:
            return next(doc for doc in documents if doc._language == language)
        except StopIteration:
            return next(doc for doc in documents if doc._language is None)


class ProductSearchViewMixin(SearchViewMixin):
    """
    Mixin class to be added to the ProductListView to restrict that list to entities matching
    the query string.
    """
    search_fields = ['product_name', 'product_code']

    def get_renderer_context(self):
        renderer_context = super().get_renderer_context()
        if renderer_context['request'].accepted_renderer.format == 'html':
            renderer_context['search_autocomplete'] = True
        return renderer_context

    def get_queryset(self):
        query = self.request.GET.get('q')
        if query:
            language = get_language_from_request(self.request)
            document = self.get_document(language)
            search = document.search().source(excludes=['body'])
            search = search.query('multi_match', query=query, fields=self.search_fields, type='bool_prefix')
            queryset = search.to_queryset()
        else:
            queryset = super().get_queryset()
        return queryset


class CatalogSearchViewMixin(SearchViewMixin):
    """
    Mixin class to be added to the ProductListView in order to create a full-text search.
    """
    search_fields = ['product_name', 'product_code', 'body']

    def get_serializer(self, *args, **kwargs):
        kwargs.setdefault('label', 'search')
        return super().get_serializer(*args, **kwargs)

    def get_queryset(self):
        language = get_language_from_request(self.request)
        document = self.get_document(language)
        query = self.request.GET.get('q')
        search = document.search().source(excludes=['body'])
        if query:
            search = search.query('multi_match', query=query, fields=self.search_fields)
        return search.to_queryset()



"""
See PEP 386 (http://www.python.org/dev/peps/pep-0386/)

Release logic:
 1. Increase version number in __version__ (below)
 2. Check that all changes have been documented in docs/changelog.rst
 3. In setup.py, assure that `classifiers` and `install_requires` reflect the latest versions.
 4. git add shop/__init__.py docs/changelog.rst setup.py
 5. git commit -m 'Bump to {new version}'
 6. git push
 7. assure that all tests pass on https://travis-ci.org/awesto/django-shop
 8. git tag {new version}
 9. git push --tags
10. python setup.py sdist
11. twine upload dist/django-shop-{new version}.tar.gz
"""
__version__ = '1.2.1'

default_app_config = 'shop.apps.ShopConfig'



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-04-11 15:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('email_auth', '0003_django110'),
    ]

    operations = []

    from django.contrib.auth import validators

    operations.append(
        migrations.AlterField(
            model_name='user',
            name='username',
            field=models.CharField(
                error_messages={'unique': 'A user with that username already exists.'},
                help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.',
                max_length=150,
                unique=True,
                validators=[validators.UnicodeUsernameValidator()],
                verbose_name='username'),
        )
    )



from django.core import checks
from django.db import models
from django.db.models.aggregates import Sum
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from shop.conf import app_settings
from shop.models.product import Availability, BaseReserveProductMixin
from shop.exceptions import ProductNotAvailable


class AvailableProductMixin:
    """
    Add this mixin class to the product models declaration, wanting to keep track on the
    current amount of products in stock. In comparison to
    :class:`shop.models.product.ReserveProductMixin`, this mixin does not reserve items in pending
    carts, with the risk for overselling. It thus is suited for products kept in the cart
    for a long period.

    The product class must implement a field named ``quantity`` accepting numerical values.
    """
    def get_availability(self, request, **kwargs):
        """
        Returns the current available quantity for this product.

        If other customers have pending carts containing this same product, the quantity
        is not not adjusted. This may result in a situation, where someone adds a product
        to the cart, but then is unable to purchase, because someone else bought it in the
        meantime.
        is not adjusted. This may result in a situation, where someone adds a product to the cart,
        but then is unable to purchase, because someone else bought it in the meantime.
        """
        def create_availability(**kwargs):
            quantity = inventory_set.aggregate(sum=Sum('quantity'))['sum']
            inventory = inventory_set.order_by('earliest').first()
            earliest = inventory.earliest
            latest = inventory_set.order_by('latest').last().latest
            if latest < now + app_settings.SHOP_LIMITED_OFFER_PERIOD:
                kwargs['limited_offer'] = True
            return Availability(quantity=quantity, earliest=earliest, latest=latest, **kwargs)

        now = timezone.now()
        inventory_set = self.inventory_set.filter(earliest__lt=now, latest__gt=now, quantity__gt=0)
        if inventory_set.exists():
            return create_availability()
        # check, if we can sell short
        later = now + app_settings.SHOP_SELL_SHORT_PERIOD
        inventory_set = self.inventory_set.filter(earliest__lt=later, latest__gt=now, quantity__gt=0)
        if inventory_set.exists():
            return create_availability(sell_short=True)
        return Availability(quantity=0)

    def deduct_from_stock(self, quantity, **kwargs):
        """
        Deduce requested quantity from all available inventories.
        """
        later = timezone.now() + app_settings.SHOP_SELL_SHORT_PERIOD
        for inventory in self.inventory_set.filter(earliest__lt=later, quantity__gt=0).order_by('earliest', 'latest'):
            reduce_by = min(inventory.quantity, quantity)
            inventory.quantity -= reduce_by
            inventory.save(update_fields=['quantity'])
            quantity -= reduce_by
            if quantity == 0:
                break
        else:
            raise ProductNotAvailable(self)

    def managed_availability(self):
        return True

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        for rel in cls._meta.related_objects:
            if rel.name == 'inventory_set':
                if rel.get_internal_type() != 'ForeignKey':
                    msg = "Field `product` in class inheriting from `BaseInventory` is not a valid foreign key pointing onto {}."
                    errors.append(checks.Error(msg.format(cls.__name__)))
                break
        else:
            msg = "A model inheriting from `BaseInventory` must implement a foreign key `product` pointing onto {}."
            errors.append(checks.Error(msg.format(cls.__name__)))
        return errors


class ReserveProductMixin(BaseReserveProductMixin, AvailableProductMixin):
    """
    Add this mixin class to the product models declaration, wanting to keep track on the
    current amount of products in stock.  In comparison to
    :class:`shop.models.product.AvailableProductMixin`, this mixin reserves items in pending
    carts, without the risk for overselling. On the other hand, the shop may run out of sellable
    items, if customers keep products in the cart for a long period, without proceeding to checkout.
    Use this mixin for products kept for a short period until checking out the cart, for
    instance for ticket sales. Ensure that pending carts are flushed regularly.
    """


class BaseInventory(models.Model):
    """
    This is a holder for the quantity of products items in stock.
    It also keeps track of the period, during which that product is available.

    The class implementing this abstract base class, must add a field named 'quantity'
    of type IntegerField, DecimalField or FloatField.
    """
    earliest = models.DateTimeField(
        _("Available after"),
        default=timezone.datetime.min.replace(tzinfo=timezone.get_current_timezone()),
        db_index=True,
    )

    latest = models.DateTimeField(
        _("Available before"),
        default=timezone.datetime.max.replace(tzinfo=timezone.get_current_timezone()),
        db_index=True,
    )

    class Meta:
        abstract = True
        verbose_name = _("Product Inventory")
        verbose_name_plural = _("Product Inventories")

    @classmethod
    def check(cls, **kwargs):
        from shop.models.cart import CartItemModel

        errors = super().check(**kwargs)
        for cart_field in CartItemModel._meta.fields:
            if cart_field.attname == 'quantity':
                break
        else:
            msg = "Class `{}` must implement a field named `quantity`."
            errors.append(checks.Error(msg.format(CartItemModel.__name__)))
        for field in cls._meta.fields:
            if field.attname == 'quantity':
                if field.get_internal_type() != cart_field.get_internal_type():
                    msg = "Field `{}.quantity` must be of same type as `{}.quantity`."
                    errors.append(checks.Error(msg.format(cls.__name__, CartItemModel.__name__)))
                break
        else:
            msg = "Class `{}` must implement a field named `quantity`."
            errors.append(checks.Error(msg.format(cls.__name__)))
        for field in cls._meta.fields:
            if field.attname == 'product_id':
                if field.get_internal_type() == 'ForeignKey':
                    if field.related_query_name() != 'inventory_set':
                        msg = "Class `{}.product` must have a related_name 'inventory_set'."
                        errors.append(checks.Error(msg.format(cls.__name__)))
                    break
                msg = "Class `{}.product` must be a foreign key pointing onto a Product model or variation of thereof."
                errors.append(checks.Error(msg.format(cls.__name__)))
        else:
            msg = "Class `{}` must implement a foreign key pointing onto a Product model or variation of thereof."
            errors.append(checks.Error(msg.format(cls.__name__)))
        return errors



from django.utils.translation import gettext_lazy as _

from cms.app_base import CMSApp


class CatalogListCMSApp(CMSApp):
    name = _("Catalog List")

    def get_urls(self, page=None, language=None, **kwargs):
        raise NotImplementedError("{} must implement method `.get_urls()`.".format(self.__class__))


class CatalogSearchApp(CMSApp):
    """
    This CMS apphook shall be used to render the list view for generic search results.
    These results are just determined by the search query and not influenced by other means,
    such as filters and categories. Usually this `Catalog Search` app is attached to a CMS
    page named "Search Results". That CMS page must be tagged with the ID: 'shop-search-product'.
    """
    name = _("Catalog Search")

    def get_urls(self, page=None, language=None, **kwargs):
        from django.conf.urls import url
        from shop.search.mixins import CatalogSearchViewMixin
        from shop.views.catalog import ProductListView

        SearchView = type('SearchView', (CatalogSearchViewMixin, ProductListView), {})
        return [
            url(r'^', SearchView.as_view(
                filter_backends=[],
                search_fields=['product_name', 'product_code', 'body']
            )),
        ]


class OrderApp(CMSApp):
    name = _("View Orders")
    cache_placeholders = False

    def get_urls(self, page=None, language=None, **kwargs):
        from django.conf.urls import url
        from shop.views.order import OrderView

        return [
            url(r'^(?P<slug>[\w-]+)/(?P<secret>[\w-]+)', OrderView.as_view(many=False)),  # publicly accessible
            url(r'^(?P<slug>[\w-]+)', OrderView.as_view(many=False)),  # requires authentication
            url(r'^', OrderView.as_view()),  # requires authentication
        ]


class PasswordResetApp(CMSApp):
    name = _("Password Reset Confirm")

    def get_urls(self, page=None, language=None, **kwargs):
        from django.conf.urls import url
        from shop.views.auth import PasswordResetConfirmView

        return [
            url(r'^(?P<uidb64>[0-9A-Za-z_\-]+)/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})',
                PasswordResetConfirmView.as_view(),
            )
        ]



from functools import reduce
import operator
from urllib.parse import urljoin

from django.apps import apps
from django.conf import settings
from django.core import cache, checks
from django.db import models
from django.db.models.aggregates import Sum
from django.db.models.functions import Coalesce
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

try:
    from django_elasticsearch_dsl.registries import registry as elasticsearch_registry
except ImportError:
    elasticsearch_registry = type('DocumentRegistry', (), {'get_documents': lambda *args: []})()

from polymorphic.managers import PolymorphicManager
from polymorphic.models import PolymorphicModel

from shop import deferred
from shop.conf import app_settings
from shop.exceptions import ProductNotAvailable


class Availability:
    """
    Contains the currently available quantity for a given product and period.
    """
    def __init__(self, **kwargs):
        """
        :param earliest:
            Point in time from when on this product will be available.

        :param latest:
            Point in time until this product will be available.

        :param quantity:
            Number of available items. The type of this value is the same as the type of ``quantity``
            in :class:`shop.models.cart.CartItemModel`.

        :param sell_short:
            If ``True``, sell the product even though it's not in stock. It then will be shipped
            at the point in time specified by ``earliest``.

        :param limited_offer:
            If ``True``, sell the product until the point in time specified by ``latest``. After
            that period, the product will not be available anymore.
        """
        tzinfo = timezone.get_current_timezone()
        self.earliest = kwargs.get('earliest', timezone.datetime.min.replace(tzinfo=tzinfo))
        self.latest = kwargs.get('latest', timezone.datetime.max.replace(tzinfo=tzinfo))
        quantity = kwargs.get('quantity', app_settings.MAX_PURCHASE_QUANTITY)
        self.quantity = min(quantity, app_settings.MAX_PURCHASE_QUANTITY)
        self.sell_short = bool(kwargs.get('sell_short', False))
        self.limited_offer = bool(kwargs.get('limited_offer', False))
        self.inventory = bool(kwargs.get('inventory', None))


class AvailableProductMixin:
    """
    Add this mixin class to the product models declaration, wanting to keep track on the
    current amount of products in stock. In comparison to
    :class:`shop.models.product.ReserveProductMixin`, this mixin does not reserve items in pending
    carts, with the risk for overselling. It thus is suited for products kept in the cart
    for a long period.

    The product class must implement a field named ``quantity`` accepting numerical values.
    """
    def get_availability(self, request, **kwargs):
        """
        Returns the current available quantity for this product.

        If other customers have pending carts containing this same product, the quantity
        is not not adjusted. This may result in a situation, where someone adds a product
        to the cart, but then is unable to purchase, because someone else bought it in the
        meantime.
        """
        return Availability(quantity=self.quantity)

    def deduct_from_stock(self, quantity, **kwargs):
        if quantity > self.quantity:
            raise ProductNotAvailable(self)
        self.quantity -= quantity
        self.save(update_fields=['quantity'])

    def managed_availability(self):
        return True

    @classmethod
    def check(cls, **kwargs):
        from shop.models.cart import CartItemModel

        errors = super().check(**kwargs)
        for cart_field in CartItemModel._meta.fields:
            if cart_field.attname == 'quantity':
                break
        else:
            msg = "Class `{}` must implement a field named `quantity`."
            errors.append(checks.Error(msg.format(CartItemModel.__name__)))
        for field in cls._meta.fields:
            if field.attname == 'quantity':
                if field.get_internal_type() != cart_field.get_internal_type():
                    msg = "Field `{}.quantity` must be of same type as `{}.quantity`."
                    errors.append(checks.Error(msg.format(cls.__name__, CartItemModel.__name__)))
                break
        else:
            msg = "Class `{}` must implement a field named `quantity`."
            errors.append(checks.Error(msg.format(cls.__name__)))
        return errors


class BaseReserveProductMixin:
    def get_availability(self, request, **kwargs):
        """
        Returns the current available quantity for this product.

        If other customers have pending carts containing this same product, the quantity
        is adjusted accordingly. Therefore make sure to invalidate carts, which were not
        converted into an order after a determined period of time. Otherwise the quantity
        returned by this function might be considerably lower, than what it could be.
        """
        from shop.models.cart import CartItemModel

        availability = super().get_availability(request, **kwargs)
        cart_items = CartItemModel.objects.filter(product=self).values('quantity')
        availability.quantity -= cart_items.aggregate(sum=Coalesce(Sum('quantity'), 0))['sum']
        return availability


class ReserveProductMixin(BaseReserveProductMixin, AvailableProductMixin):
    """
    Add this mixin class to the product models declaration, wanting to keep track on the
    current amount of products in stock. In comparison to
    :class:`shop.models.product.AvailableProductMixin`, this mixin reserves items in pending
    carts, without the risk for overselling. On the other hand, the shop may run out of sellable
    items, if customers keep products in the cart for a long period, without proceeding to checkout.
    Use this mixin for products kept for a short period until checking out the cart, for
    instance for ticket sales. Ensure that pending carts are flushed regularly.

    The product class must implement a field named ``quantity`` accepting numerical values.
    """


class BaseProductManager(PolymorphicManager):
    """
    A base ModelManager for all non-object manipulation needs, mostly statistics and querying.
    """
    def select_lookup(self, search_term):
        """
        Returning a queryset containing the products matching the declared lookup fields together
        with the given search term. Each product can define its own lookup fields using the
        member list or tuple `lookup_fields`.
        """
        filter_by_term = (models.Q((sf, search_term)) for sf in self.model.lookup_fields)
        queryset = self.get_queryset().filter(reduce(operator.or_, filter_by_term))
        return queryset

    def indexable(self):
        """
        Return a queryset of indexable Products.
        """
        queryset = self.get_queryset().filter(active=True)
        return queryset


class PolymorphicProductMetaclass(deferred.PolymorphicForeignKeyBuilder):

    @classmethod
    def perform_meta_model_check(cls, Model):
        """
        Perform some safety checks on the ProductModel being created.
        """
        if not isinstance(Model.objects, BaseProductManager):
            msg = "Class `{}.objects` must provide ModelManager inheriting from BaseProductManager"
            raise NotImplementedError(msg.format(Model.__name__))

        if not isinstance(getattr(Model, 'lookup_fields', None), (list, tuple)):
            msg = "Class `{}` must provide a tuple of `lookup_fields` so that we can easily lookup for Products"
            raise NotImplementedError(msg.format(Model.__name__))

        if not callable(getattr(Model, 'get_price', None)):
            msg = "Class `{}` must provide a method implementing `get_price(request)`"
            raise NotImplementedError(msg.format(cls.__name__))


class BaseProduct(PolymorphicModel, metaclass=PolymorphicProductMetaclass):
    """
    An abstract basic product model for the shop. It is intended to be overridden by one or
    more polymorphic models, adding all the fields and relations, required to describe this
    type of product.

    Some attributes for this class are mandatory. They shall be implemented as property method.
    The following fields MUST be implemented by the inheriting class:
    ``product_name``: Return the pronounced name for this product in its localized language.

    Additionally the inheriting class MUST implement the following methods ``get_absolute_url()``
    and ``get_price()``. See below for details.

    Unless each product variant offers its own product code, it is strongly recommended to add
    a field ``product_code = models.CharField(_("Product code"), max_length=255, unique=True)``
    to the class implementing the product.
    """
    created_at = models.DateTimeField(
        _("Created at"),
        auto_now_add=True,
    )

    updated_at = models.DateTimeField(
        _("Updated at"),
        auto_now=True,
    )

    active = models.BooleanField(
        _("Active"),
        default=True,
        help_text=_("Is this product publicly visible."),
    )

    class Meta:
        abstract = True
        verbose_name = _("Product")
        verbose_name_plural = _("Products")

    def product_type(self):
        """
        Returns the polymorphic type of the product.
        """
        return force_str(self.polymorphic_ctype)
    product_type.short_description = _("Product type")

    @property
    def product_model(self):
        """
        Returns the polymorphic model name of the product's class.
        """
        return self.polymorphic_ctype.model

    def get_absolute_url(self):
        """
        Hook for returning the canonical Django URL of this product.
        """
        msg = "Method get_absolute_url() must be implemented by subclass: `{}`"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def get_price(self, request):
        """
        Hook for returning the current price of this product.
        The price shall be of type Money. Read the appropriate section on how to create a Money
        type for the chosen currency.
        Use the `request` object to vary the price according to the logged in user,
        its country code or the language.
        """
        msg = "Method get_price() must be implemented by subclass: `{}`"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def get_product_variant(self, **kwargs):
        """
        Hook for returning the variant of a product using parameters passed in by **kwargs.
        If the product has no variants, then return the product itself.

        :param **kwargs: A dictionary describing the product's variations.
        """
        return self

    def get_product_variants(self):
        """
        Hook for returning a queryset of variants for the given product.
        If the product has no variants, then the queryset contains just itself.
        """
        return self._meta.model.objects.filter(pk=self.pk)

    def get_availability(self, request, **kwargs):
        """
        Hook for checking the availability of a product.

        :param request:
            Optionally used to vary the availability according to the logged in user,
            its country code or language.

        :param **kwargs:
            Extra arguments passed to the underlying method. Useful for products with
            variations.

        :return: An object of type :class:`shop.models.product.Availability`.
        """
        return Availability()

    def managed_availability(self):
        """
        :return True: If this product has its quantity managed by some inventory functionality.
        """
        return False

    def is_in_cart(self, cart, watched=False, **kwargs):
        """
        Checks if the current product is already in the given cart, and if so, returns the
        corresponding cart_item.

        :param watched (bool): This is used to determine if this check shall only be performed
            for the watch-list.

        :param **kwargs: Optionally one may pass arbitrary information about the product being looked
            up. This can be used to determine if a product with variations shall be considered
            equal to the same cart item, resulting in an increase of it's quantity, or if it
            shall be considered as a separate cart item, resulting in the creation of a new item.

        :returns: The cart item (of type CartItem) containing the product considered as equal to the
            current one, or ``None`` if no product matches in the cart.
        """
        from shop.models.cart import CartItemModel
        cart_item_qs = CartItemModel.objects.filter(cart=cart, product=self)
        return cart_item_qs.first()

    def deduct_from_stock(self, quantity, **kwargs):
        """
        Hook to deduct a number of items of the current product from the stock's inventory.

        :param quantity: Number of items to deduct.

        :param **kwargs:
            Extra arguments passed to the underlying method. Useful for products with
            variations.
        """

    def get_weight(self):
        """
        Optional hook to return the product's gross weight in kg. This information is required to
        estimate the shipping costs. The merchants product model shall override this method.
        """
        return 0

    @classmethod
    def check(cls, **kwargs):
        """
        Internal method to check consistency of Product model declaration on bootstrapping
        application.
        """
        errors = super().check(**kwargs)
        try:
            cls.product_name
        except AttributeError:
            msg = "Class `{}` must provide a model field implementing `product_name`"
            errors.append(checks.Error(msg.format(cls.__name__)))
        return errors

    def update_search_index(self):
        """
        Update the Document inside the Elasticsearch index after changing relevant parts
        of the product.
        """
        documents = elasticsearch_registry.get_documents([ProductModel])
        if settings.USE_I18N:
            for language, _ in settings.LANGUAGES:
                try:
                    document = next(doc for doc in documents if doc._language == language)
                except StopIteration:
                    document = next(doc for doc in documents if doc._language is None)
                document().update(self)
        else:
            document = next(doc for doc in documents)
            document().update(self)

    def invalidate_cache(self):
        """
        Method ``ProductCommonSerializer.render_html()`` caches the rendered HTML snippets.
        Invalidate this HTML snippet after changing relevant parts of the product.
        """
        shop_app = apps.get_app_config('shop')
        if shop_app.cache_supporting_wildcard:
            cache.delete_pattern('product:{}|*'.format(self.id))

ProductModel = deferred.MaterializedModel(BaseProduct)


class CMSPageReferenceMixin:
    """
    Products which refer to CMS pages in order to emulate categories, normally need a method for
    being accessed directly through a canonical URL. Add this mixin class for adding a
    ``get_absolute_url()`` method to any to product model.
    """
    category_fields = ['cms_pages']  # used by ProductIndex to fill the categories

    def get_absolute_url(self):
        """
        Return the absolute URL of a product
        """
        # sorting by highest level, so that the canonical URL
        # associates with the most generic category
        cms_page = self.cms_pages.order_by('node__path').last()
        if cms_page is None:
            return urljoin('/category-not-assigned/', self.slug)
        return urljoin(cms_page.get_absolute_url(), self.slug)



from django.core.exceptions import ValidationError
from django.forms import fields, widgets
from django.template import engines
from django.template.loader import select_template
from django.utils.translation import gettext_lazy as _, gettext
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from shop.cascade.plugin_base import ShopPluginBase
from shop.conf import app_settings


class ShopSearchResultsFormMixin(EntangledModelFormMixin):
    CHOICES = [
        ('paginator', _("Use Paginator")),
        ('manual', _("Manual Infinite")),
        ('auto', _("Auto Infinite")),
    ]

    pagination = fields.ChoiceField(
        label=_("Pagination"),
        choices=CHOICES,
        widget=widgets.RadioSelect,
        help_text=_("Shall the list of search results use a paginator or scroll infinitely?"),
    )

    class Meta:
        entangled_fields = {'glossary': ['pagination']}

    def clean(self):
        cleaned_data = super().clean()
        page = self.instance.placeholder.page if self.instance.placeholder_id else None
        if page and page.application_urls != 'CatalogSearchApp':
            raise ValidationError("This plugin can only be used on a CMS page with an application of type 'Search'.")
        return cleaned_data


class ShopSearchResultsPlugin(ShopPluginBase):
    name = _("Search Results")
    require_parent = True
    parent_classes = ['BootstrapColumnPlugin']
    form = ShopSearchResultsFormMixin
    cache = False

    def get_render_template(self, context, instance, placeholder):
        if instance.placeholder.page.application_urls != 'CatalogSearchApp':
            alert_msg = '''<div class="alert alert-danger">
            Plugin "{}" is used on a CMS page without an application of type "Catalog Search".
            </div>'''
            return engines['django'].from_string(alert_msg.format(self.name))
        return select_template([
            '{}/search/results.html'.format(app_settings.APP_LABEL),
            'shop/search/results.html',
        ])

    def render(self, context, instance, placeholder):
        self.super(ShopSearchResultsPlugin, self).render(context, instance, placeholder)
        context['pagination'] = instance.glossary.get('pagination', 'paginator')
        return context

    @classmethod
    def get_identifier(cls, obj):
        pagination = obj.glossary.get('pagination')
        if pagination == 'paginator':
            return gettext("Manual Pagination")
        return gettext("Infinite Scroll")

plugin_pool.register_plugin(ShopSearchResultsPlugin)



from shop.conf import app_settings
from shop.models.customer import CustomerModel, VisitingCustomer


def customer(request):
    """
    Add the customer to the RequestContext
    """
    msg = "The request object does not contain a customer. Edit your MIDDLEWARE_CLASSES setting to insert 'shop.middlerware.CustomerMiddleware'."
    assert hasattr(request, 'customer'), msg

    customer = request.customer
    if request.user.is_staff:
        try:
            customer = CustomerModel.objects.get(pk=request.session['emulate_user_id'])
        except CustomerModel.DoesNotExist:
            customer = VisitingCustomer()
        except (AttributeError, KeyError):
            pass
    return {'customer': customer}


def shop_settings(request):
    """
    Add configuration settings to the context to customize the shop's settings in templates
    """
    from rest_auth.app_settings import LoginSerializer
    return {
        'site_header': app_settings.APP_LABEL.capitalize(),
        'EDITCART_NG_MODEL_OPTIONS': app_settings.EDITCART_NG_MODEL_OPTIONS,
        'ADD2CART_NG_MODEL_OPTIONS': app_settings.ADD2CART_NG_MODEL_OPTIONS,
        'ALLOW_SHORT_SESSIONS': 'stay_logged_in' in LoginSerializer().fields,
        'LINK_TO_EMPTY_CART': app_settings.LINK_TO_EMPTY_CART,
    }



import warnings

from django.apps import AppConfig
from django.core.cache import cache
from django.utils.translation import gettext_lazy as _


class ShopConfig(AppConfig):
    name = 'shop'
    verbose_name = _("Shop")
    cache_supporting_wildcard = False

    def ready(self):
        from rest_framework.serializers import ModelSerializer
        from shop.deferred import ForeignKeyBuilder
        from shop.models.fields import JSONField
        from shop.rest.fields import JSONSerializerField
        from shop.patches import PageAttribute
        from cms.templatetags import cms_tags

        # add JSONField to the map of customized serializers
        ModelSerializer.serializer_field_mapping[JSONField] = JSONSerializerField

        # perform some sanity checks
        ForeignKeyBuilder.check_for_pending_mappings()

        cms_tags.register.tags['page_attribute'] = PageAttribute

        if callable(getattr(cache, 'delete_pattern', None)):
            self.cache_supporting_wildcard = True
        else:
            warnings.warn("\n"
                "Your caching backend does not support invalidation by key pattern.\n"
                "Please use `django-redis-cache`, or wait until the product's HTML\n"
                "snippet cache expires by itself.")
